/*
 * Radius information module for icradius.
 * Version 3.0
 * for apache 1.3 [ conversion in process... ]
 *       apache 2.0 module working.
 */
/* 
 * Radwho Module.  Display configuration information for the server and
 * all included modules.
 *
 * <Location /online>
 * SetHandler radwho
 * AuthType   Basic
 * AuthName   Radius
 * Require    valid-user
 * RADAuthFile /etc/radwho    // File containing the username/password pairs used for radwho authentication
 * RADUserLimit 57600           // Seconds used before a login is highlighted
 * RADDBUser	User	// Database username
 * RADSortBy   NASPortID     // Default sorting of entries in the table.  Case sensitive.
 * RADDBPass  MyPassword    // MySQL Login Password
 * RADServer  localhost          // MySQL server host
 * RADDatabase radius            // The MySQL database, on the host, that contains the login information
 * RADURL http://www.example.org/radwho/	// Full URI of this module for self-referencing URL calls	
 * RADHighlight "#ff9999"	// Color [in html format so quote it] to highlight abusive logins.
 * </Location>
 *
 * GET /online - Returns who is currently online
 * GET /online/clear?RadAcctId  - deletes the accounting record with the given Id
 * GET /online/data?name=?????  - gets radgroup data and logins for the day
 * GET /online/today            - lists everybody who has logged in today.
 * GET /online/sort?by=????     - sorts the table according to the selected critiria
 * GET /online/showme?id        - shows the entire 'id' record in the database.
 * GET /online/todays?days=n    - Show all logins for the past n days
 *
 * Robert 'Admiral' Coeyman <admiral@corner.net> 2001
 *
 * In 2002 
 *  Added the authentication routines
 *  Cleaned up the look of the page.
 *  Set the refresh in the header.
 *  Created the clearing function for 'stuck' logins.
 *  Made "/clear" case insensitive.
 *  Added the data directive.
 *  Added code to hypothesize if the radius server is running or now.	
 *  Wrote a routine to show the entire login record for a current login.	
 *  Changed the clear instruction to create a logout with an
 *   AcctTerminateCause of 'Dead Login' and a session length of 5 seconds.
 *  [ Otherwise, our synchronization daemon would restore the dead login. ]     
 *
 * May 2002 by Isam Ishaq <isam@planet.edu>  *  Added /sort directive.
 *  Added CallingStationID "Calling From" to the /data table.
 *  Clicking on the user name shows the logins by that user.
 *
 *
 *  Rewrite begins in 2005.
 *   Add support for Apache 2.0 and allow both versions to take
 *   Configuration options from the httpd.conf file.
 * The configuration directives are limited to the <Directory><Location> and <Files> sections
 * of the main apache configuration file.  htaccess overrides are not allowed.
 */
/* Config stuff....
 * MODULE-DEFINITION-START
 * Name: radwho_module
 * ConfigStart
echo "Make sure that libmysqlclient.a is where apache can find it" 
LIBS="$LIBS -L/usr/local/lib/mysql -lgdbm -lmysqlclient -lm"
 * ConfigEnd
 * MODULE-DEFINITION-END
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
#include "http_conf_globals.h"
#include <stdio.h>
#include <time.h>
//#include "radwho.h"
#include <mysql/mysql.h>
#include <gdbm.h>          /* database header.        */
#include <string.h>        /* for strncmp,strstr      */
#include <unistd.h>

// #define WANT_APR_STRFUNC 
#define MODULE_RELEASE "3.0.1"
// #define CORE_PRIVATE

/*--------------------------------------------------------------------------*/ 
/* */ 
/* Data declarations.  */ 
/* */ 
/* Here are the static cells and structure declarations private to our */ 
/* module.  */ 
/* */ 
/*--------------------------------------------------------------------------*/

extern char **environ;

/*
 * Radius configuration record.  Used for both per-directory and per-server
 * configuration data.
 *
 */ 
typedef struct rad_cfg {
	char *RauthFile; 	/* gdmb file 		*/
	char *Rdbuser;  	/* for mysql login 	*/
	char *Rdbpass;  	/* for mysql login 	*/
	char *Rdatabase; 	/* For Mysql Login 	*/
	char *Rurlpath;  	/* to pass the url 	*/
	char *Rhcolor;  	/* Highlight color 	*/
	char *RLimit;  		/* Limit seconds 	*/
	char *RSortBy;  	/* Sortby setting	*/ 
	char *Rdbserver;  	/* For MYSQL login	*/ 
} rad_cfg;

typedef struct sql_socket 
{
	MYSQL *sock;
	MYSQL conn;
	MYSQL_RES *result;
	MYSQL_RES *result2;
	MYSQL_RES *result3; 
} SQLSOCK;

#define NULL ((void *)0)

//module AP_MODULE_DECLARE_DATA radwho_module;

/* Try making this global */
	SQLSOCK *sockit;

/* Data Structures for the module */
module MODULE_VAR_EXPORT radwho_module;

/*
 * Declare ourselves so the configuration routines can find and know us.
 * We'll fill it in at the end of the module.
 */ 


static void *create_radwho_config_record(pool *p, char *d ) 
{
	/* create the config record. */
	rad_cfg *conf=(rad_cfg *)ap_pcalloc(p,sizeof(rad_cfg));
	if ( conf == NULL )
		{
		syslog(LOG_ERR,"Radwho got no memory.");
		return NULL;
		}

	/* fill in null values */
/* Null values crash the system */
	conf->RauthFile = "/etc/apache/radwho";
	conf->Rdbuser = "icradius";
	conf->Rdbpass = NULL;
	conf->Rdatabase = "icradius";
	conf->Rurlpath = NULL;
	conf->Rhcolor = NULL;
	conf->RLimit = "63360";
	conf->RSortBy = "NASPortId";
	conf->Rdbserver = "localhost";

/* done--return with the configuration block */
		syslog(LOG_INFO,"Creating Config DONE");

	return conf; 
}

/* ------------------------------------------------------------------
 * The data setting functions
 *
 * ------------------------------------------------------------------
 */

/* Converted for apache 1.3 */
static const char *set_rad_limitflag( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	syslog(LOG_INFO,"Get Conf Record.");
	/* get the conf record */
	conf=(rad_cfg *)dir_config;

	syslog(LOG_INFO,"Copy Limit flag to Conf structure.");
	/* load the data */
	conf->RLimit = (char *)ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set Limit Flag to %s.",arg);

	/*return NULL */
	return NULL; 

}


/* Converted for apache 1.3 */
static const char *set_rad_sortby( cmd_parms *cmd, void *dir_config,
					char *arg ) 
{
	rad_cfg *conf;


	if ( arg == NULL ) {
		syslog(LOG_ERR,"SortBy set to NULL");
		return NULL;
	}

	syslog(LOG_INFO,"Get Conf Record.");
	/* get the conf record */
	conf=(rad_cfg *)dir_config;

	if ( conf == NULL ) {
		syslog(LOG_ERR,"COnfig record has NULL pointer?");
		return NULL;
	}

	syslog(LOG_INFO,"Set Sortby");
	/* load the data */
	conf->RSortBy = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set Sortby to %s",arg);

	/*return NULL */
	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_dbuser( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	/* get the record */
	conf=(rad_cfg *)dir_config;

	/* load the data */
	conf->Rdbuser = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set DB User.");

	/*return NULL */
	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_dbpass( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	/* get the record */
	conf=(rad_cfg *)dir_config;


	/* load the data */
	conf->Rdbpass = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set DB Password.");
	/*return NULL */
	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_database( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	/* get the record */
	conf=(rad_cfg *)dir_config;

	/* load the data */
	conf->Rdatabase = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set Database to %s",arg);
	/*return NULL */
	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_urlpath( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	/* get the record */
	conf=(rad_cfg *)dir_config;

	/* load the data */
	conf->Rurlpath = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set URL Path to %s.",arg);
	/*return NULL */
	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_highlight( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	/* get the record */
	conf=(rad_cfg *)dir_config;

	/* load the data */
	conf->Rhcolor = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set Highlight to %s.",arg);
	/*return NULL */

	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_dbserver( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	/* get the record */
	conf=(rad_cfg *)dir_config;

	/* load the data */
	conf->Rdbserver = ap_pstrdup(cmd->pool,arg);

	syslog(LOG_INFO,"Set DB Server to %s.",arg);
	/*return NULL */

	return NULL; 
}

/* Converted for apache 1.3 */
static const char *set_rad_authfile( cmd_parms *cmd, void *dir_config,
					const char *arg ) 
{
	rad_cfg *conf;

	syslog(LOG_INFO,"Get config record.");
	/* get the record */
	conf=(rad_cfg *)dir_config;

	/* load the data */
	conf->RauthFile = (char *)ap_pstrdup(cmd->pool,arg);
	syslog(LOG_INFO,"Set AuthFile to %s.",arg);

	/*return NULL */

	return NULL; 
}


/* -----------------------------------------------------------------
 * start the internal routines
 *
 * ------------------------------------------------------------------
 */ 

/* function to print out records from an apr_table */
int writeout_func( void *rec, const char *key, const char *value) 
{
	request_rec *r = (request_rec *)rec;

/* Stop if the key is null or if there is no value to print */ 

if ( key == NULL || value == NULL || value[0] == '\0') return 1;

/* write the data in a 2 element table */
    ap_rprintf(r,"<tr><td>%s</td><td>%s</td></tr>\n",key,value);

/* return success */ 
return 1; 
}

/* this routine shows who used an IP at a specific time. */ 
static int rad_whohad( rad_cfg *conf, request_rec *r, SQLSOCK *sock) 
{
	char *p,*pp,*ppp,buf[256];
	char *ip,*year,*month,*day,*hour,*min,*sec;
	MYSQL_ROW row;


	p=r->args;
	pp=strstr(p,"=");
	ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
ip=pp;
	pp=strstr(ppp,"=");
	p =strstr(ppp,"&");
if ( p != NULL ) *p=(char) 0;
if ( p != NULL ) p++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
year=pp;
	pp=strstr(p,"=");
	ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
month=pp;
	pp=strstr(ppp,"=");
	p =strstr(ppp,"&");
if ( p != NULL ) *p=(char) 0;
if ( p != NULL ) p++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
day=pp;
	pp=strstr(p,"=");
	ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
hour=pp;
	pp=strstr(ppp,"=");
	p =strstr(ppp,"&");
if ( p != NULL ) *p=(char) 0;
if ( p != NULL ) p++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
min=pp;
	pp=strstr(p,"=");
	ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
sec=pp;

		sprintf((char *)&buf,"SELECT * FROM radacct WHERE FramedIPAddress='%s' AND AcctStartTime <='%s-%s-%s %s:%s:%s' AND AcctStopTime >='%s-%s-%s %s:%s:%s'",ip,
		year,month,day,hour,min,sec,
		year,month,day,hour,min,sec);
	/* Trap Errors first */
		if (sockit->sock == NULL ) {
			ap_rputs("<B>Failed to QUERY the radius server</B>",r);
			ap_rprintf(r,"<BR>Error: No Socket<BR></BODY></HTML>\n");
			return 0;
		} /* end of the error Trap */
		mysql_query(sockit->sock,(char *)&buf);
		syslog(LOG_INFO,"Asking about %s",buf);
		sockit->result=mysql_store_result(sockit->sock);
		if ( sockit->result == NULL ) {
			ap_rputs("<b>Unable to search database.</b><br><br>",r);
			return 0;
		}

		row=NULL;

		if ( sockit->result != NULL) { 
			row = mysql_fetch_row(sockit->result);
		} else {
			ap_rprintf(r,"NOBODY had %s at %s:%s:%s on %s-%s-%s\n<br><br>",ip,
			hour,min,sec,year,month,day);
			return 1;
		}

			if ( mysql_num_rows(sockit->result) < 1 ) { 
				ap_rprintf(r,"NOBODY was on %s at %s:%s:%s on %s-%s-%s\n<br><br>",ip,
				hour,min,sec,year,month,day);
				return 1;
			}

			if ( row == NULL ) {
				syslog(LOG_ERR,"ERROR: Exposing login %s [%s]...",r->args,buf);
				ap_rprintf(r,"NOBODY had %s at %s:%s:%s on %s-%s-%s !\n<br><br>",ip,
					hour,min,sec,year,month,day);
				return 0;
			}
			syslog(LOG_INFO,"Exposing login [%s]...",buf);
			ap_rputs("<table width=57% bgcolor=\"white\" border=2>\n",r);
			ap_rputs("<tr><th>Accounting ID</th>\n",r);
			if ( row[0] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[0]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[0]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Session ID</th>\n",r);
			if ( row[1] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[1]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[1]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>User Name</th>\n",r);
			if ( row[2] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[2]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[2]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Realm</th>\n",r);
			if ( row[3] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[3]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[3]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>NAS IP Address</th>\n",r);
			if ( row[4] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[4]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[4]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Port</th>\n",r);
			if ( row[5] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[5]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[5]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Port Type</th>\n",r);
			if ( row[6] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[6]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[6]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Login Time</th>\n",r);
			if ( row[7] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[7]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[7]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Logout Time</th>\n",r);
			if ( row[8] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[8]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[8]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Session Time</th>\n",r);
			if ( row[9] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[9]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[9]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Authenticated By</th>\n",r);
			if ( row[10] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[10]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[10]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Connection Info</th>\n",r);
			if ( row[11] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[11]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[11]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Input Bytes</th>\n",r);
			if ( row[12] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[12]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[12]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Output Bytes</th>\n",r);
			if ( row[13] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[13]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[13]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Called Station ID</th>\n",r);
			if ( row[14] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[14]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[14]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Calling Station ID</th>\n",r);
			if ( row[15] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[15]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[15]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Termination Cause</th>\n",r);
			if ( row[16] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[16]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[16]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Service Type</th>\n",r);
			if ( row[17] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[17]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[17]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Protocol</th>\n",r);
			if ( row[18] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[18]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[18]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>IP Address</th>\n",r);
			if ( row[19] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[19]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[19]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Start Delay</th>\n",r);
			if ( row[20] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[20]);
				syslog(LOG_INFO,"Exposing login [%s]...",row[20]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
				return 0;
			}
			ap_rputs("<tr><th>Stop Delay</th>\n",r);
			if ( row[21] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[21]);
				syslog(LOG_INFO,"Exposing login [%s] EOF",row[21]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
/* done with the result--free the data */
	
	if ( sockit->result != NULL) mysql_free_result(sockit->result); 
	sockit->result=NULL;
	
return 1;
} /* end of the who-had function */

/* function to clear dead logins. */

static int rad_clear( rad_cfg *conf, request_rec *r, SQLSOCK *sock) 
{
        char buf[256];

/* Function called inside the table, so frame output */
ap_rputs("<tr><th colspan=7 bgcolor=\"#f6debc\">\n",r); 

		sprintf((char *)&buf,"UPDATE radacct SET AcctStopTime=NOW(),AcctSessionTime='5',AcctTerminateCause='Dead Login' WHERE RadAcctID = '%s'",r->args);
	/* Trap Errors first */
		if (sock->sock == NULL ) {
			ap_rputs("<B>Failed to QUERY the radius server</B>",r);
			ap_rprintf(r,"<BR>Error: No Socket<BR></th></tr>\n");
			if ( sock->result != NULL) mysql_free_result(sock->result);
			return 0;
		} /* end of the error Trap */
		mysql_query(sock->sock,(char *)&buf);
		syslog(LOG_INFO,"Dead login '%s' removed '%s'",r->args,mysql_error(sock->sock));
		//if ( sock->sock != NULL) ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"Mysql reply %s",mysql_error(sock->sock));
		ap_rputs("<I>Dead login removed successfully.</I>",r);
		ap_rprintf(r," Login ID:<b>%s</b></th></tr>\n",r->args);
return 1;
} /* end of the clearing instruction */


/* Function to show a single login record */
static int rad_showme( rad_cfg *conf, request_rec *r, SQLSOCK *sock) 
{
	int iii,res,days,extra;
	MYSQL_ROW row,brow;
	MYSQL_RES *localres;
        char buf[256], string[128], *p,ports[49];


		sprintf((char *)&buf,"SELECT * FROM radacct WHERE RadAcctID = '%s'",r->args);
	/* Trap Errors first */
		if (sockit->sock == NULL ) {
			ap_rputs("<B>Failed to QUERY the radius server</B>",r);
			ap_rprintf(r,"<BR>Error: No Socket<BR></BODY></HTML>\n");
			if (sockit->result != NULL) mysql_free_result(sockit->result);
			return 0;
		} /* end of the error Trap */
		if (sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
		syslog(LOG_INFO,"Exposing login '%s' ...",r->args);
		if (sockit->sock != NULL) localres=mysql_store_result(sockit->sock);

/* Not ALL of this data is available for users still online, however,
  	this command doesn't have to only be run for them... */

		if ( localres == NULL ) {
			ap_rputs("<B>ERROR: Unable to find that login online.</B><br>\n",r);
			if (localres != NULL) mysql_free_result(localres);
		} else { /* got data--print it out */
			if ( localres != NULL) { 
				row = mysql_fetch_row(localres);
			} else {
				return 0;
			}
			if ( row == NULL ) {
				syslog(LOG_ERR,"ERROR:Exposing login %s [%s]...",r->args,buf);
				return 0;
			}
			ap_rputs("<table width=57% bgcolor=\"white\" border=2>\n",r);
			ap_rputs("<tr><th>Accounting ID</th>\n",r);
			if ( row[0] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[0]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Session ID</th>\n",r);
			if ( row[1] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[1]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>User Name</th>\n",r);
			if ( row[2] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[2]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Realm</th>\n",r);
			if ( row[3] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[3]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>NAS IP Address</th>\n",r);
			if ( row[4] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[4]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Port</th>\n",r);
			if ( row[5] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[5]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Port Type</th>\n",r);
			if ( row[6] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[6]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Login Time</th>\n",r);
			if ( row[7] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[7]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Logout Time</th>\n",r);
			if ( row[8] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[8]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Session Time</th>\n",r);
			if ( row[9] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[9]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Authenticated By</th>\n",r);
			if ( row[10] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[10]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Connection Info</th>\n",r);
			if ( row[11] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[11]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Input Bytes</th>\n",r);
			if ( row[12] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[12]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Output Bytes</th>\n",r);
			if ( row[13] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[13]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Called Station ID</th>\n",r);
			if ( row[14] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[14]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Calling Station ID</th>\n",r);
			if ( row[15] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[15]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Termination Cause</th>\n",r);
			if ( row[16] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[16]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Service Type</th>\n",r);
			if ( row[17] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[17]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Protocol</th>\n",r);
			if ( row[18] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[18]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>IP Address</th>\n",r);
			if ( row[19] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[19]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Start Delay</th>\n",r);
			if ( row[20] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[20]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Stop Delay</th>\n",r);
			if ( row[21] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s </td></tr>\n",row[21]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}

/* now get averaged data for this user. */
		sprintf((char *)&buf,"SELECT AVG(AcctSessionTime),MIN(AcctSessionTime),MAX(AcctSessionTime),SUM(AcctSessionTime),STD(AcctSessionTime) FROM radacct WHERE UserName = '%s' AND AcctSessionTime > '0'",row[2]);
		if (sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
		syslog(LOG_INFO,"Exposing login %s ...",r->args);
		if (sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock);

		if ( sockit->result2 != NULL ) {
				brow=mysql_fetch_row(sockit->result2);
			} else {
				brow = NULL;
		}


		if ( brow != NULL ) {
			ap_rputs("<tr><th>Average Login</th>\n",r);
			if ( brow[0] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s Seconds </td></tr>\n",brow[0]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Minimum Login</th>\n",r);
			if ( brow[1] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s seconds </td></tr>\n",brow[1]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Maximum Login</th>\n",r);
			if ( brow[2] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s seconds </td></tr>\n",brow[2]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Total [known] Login Time</th>\n",r);
			if ( brow[3] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s seconds </td></tr>\n",brow[3]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
			ap_rputs("<tr><th>Standard Deviation</th>\n",r);
			if ( brow[4] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s seconds</td></tr>\n",brow[4]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}
		      } /* end of the NULL brow trap */
			ap_rputs("</table><br><br><br>\n\n",r);
			if (sockit->result2 != NULL) mysql_free_result(sockit->result2);


/* Show all logins for the last week */
if ( row[2] != NULL ) {
		sprintf((char *)&buf,"SELECT RadAcctID,AcctStartTime,AcctStopTime,AcctTerminateCause,AcctSessionTime FROM radacct WHERE TO_DAYS(AcctStartTime >= 'TO_DAYS(NOW())-7' AND Username = '%s'",row[2]);
		if ( sockit->sock != NULL ) {
			if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
			if ( sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock);
			if ( sockit->result != NULL ) {
			if ( sockit->result2 != NULL){
				res = mysql_num_rows(sockit->result2); 
			} else {
				res=0;
			}
				if ( res > 0 ) {
				ap_rputs("<table width=100% border=1 bgcolor=\"#ffffff\">\n",r);
				ap_rputs("<tr><td>ID </td>\n",r);
				ap_rputs("<td>Log On</td>\n",r);
				ap_rputs("<td>Log Off</td>\n",r);
				ap_rputs("<td>Termination</td>\n",r);
				ap_rputs("<td>Login Length</td></tr>\n",r);


			for ( iii=0;iii<res;iii++) { 
					brow=mysql_fetch_row(sockit->result2);
			if ( brow[0] != NULL ) {
				ap_rprintf(r,"<tr><td><a href=\"%s/showme?%s\">&nbsp %s</a></td>\n",conf->Rurlpath,brow[0],brow[0]);
				} else {
				ap_rputs("<tr><td>&nbsp N/A </td>\n",r);
			}
			if ( brow[1] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s</td>\n",brow[1]);
				} else {
				ap_rputs("<td>&nbsp N/A </td>\n",r);
			}
			if ( brow[2] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s</td>\n",brow[2]);
				} else {
				ap_rputs("<td>&nbsp N/A </td>\n",r);
			}
			if ( brow[3] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s</td>\n",brow[3]);
				} else {
				ap_rputs("<td>&nbsp N/A </td>\n",r);
			}
			if ( brow[4] != NULL ) {
				ap_rprintf(r,"<td>&nbsp %s</td></tr>\n",brow[4]);
				} else {
				ap_rputs("<td>&nbsp N/A </td></tr>\n",r);
			}

			} /* end of the counter loop */
				ap_rputs("</table><br>\n",r);
				} /* no user trap */
			} /* end of the have data to talk about */
		} /* end of have socket */
} /* End of have username */
			if (localres  != NULL) mysql_free_result(localres);
			if ( sockit->result2 != NULL) mysql_free_result(sockit->result2);
		} /* end of the printout. */
return 1;
} /* end of the showme request. */

/* function to print a single day's login data */
static int rad_today( rad_cfg *conf, request_rec *r, SQLSOCK *sock) 
{
	int iii,comma,pct,ii,xxx;
	struct tm *btime,*today;
	time_t now,then,online, limit;
	MYSQL_ROW row,brow;
        char buf[256], string[128], *p,ports[49], sortby[50];


		/* start by creating my table */
ap_rputs("<BR><TABLE WIDTH=100% BORDER=1 BGCOLOR=\"ffffff\">",r);
ap_rputs("<TR><TH>UserName:</TH><TH>0<BR>0</TH><TH>0<BR>1</TH><TH>0<BR>2</TH>",r);
ap_rputs("<TH>0<BR>3</TH><TH>0<BR>4</TH><TH>0<BR>5</TH><TH>0<BR>6</TH>",r);
ap_rputs("<TH>0<BR>7</TH><TH>0<BR>8</TH><TH>0<BR>9</TH><TH>1<BR>0</TH>",r);
ap_rputs("<TH>1<BR>1</TH><TH>1<BR>2</TH><TH>1<BR>3</TH><TH>1<BR>4</TH>",r);
ap_rputs("<TH>1<BR>5</TH><TH>1<BR>6</TH><TH>1<BR>7</TH><TH>1<BR>8</TH>",r);
ap_rputs("<TH>1<BR>9</TH><TH>2<BR>0</TH><TH>2<BR>1</TH><TH>2<BR>2</TH>",r);
ap_rputs("<TH>2<BR>3</TH><TH>2<BR>4</TH><TH>2<BR>5</TH><TH>2<BR>6</TH>",r);
ap_rputs("<TH>2<BR>7</TH><TH>2<BR>7</TH><TH>2<BR>9</TH><TH>3<BR>0</TH>",r);
ap_rputs("<TH>3<BR>1</TH><TH>3<BR>2</TH><TH>3<BR>3</TH><TH>3<BR>4</TH>",r);
ap_rputs("<TH>3<BR>5</TH><TH>3<BR>6</TH><TH>3<BR>7</TH><TH>3<BR>8</TH>",r);
ap_rputs("<TH>3<BR>9</TH><TH>4<BR>0</TH><TH>4<BR>1</TH><TH>4<BR>2</TH>",r);
ap_rputs("<TH>4<BR>3</TH><TH>4<BR>4</TH><TH>4<BR>5</TH><TH>4<BR>6</TH>",r);
ap_rputs("<TH>4<BR>7</TH><TH>4<BR>8</TH></TR>",r); 

/* run the query */
	sprintf((char *)&buf,"SELECT DISTINCT UserName FROM radacct WHERE TO_DAYS(AcctStartTime) = TO_DAYS(NOW()) ORDER BY UserName");
	if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);

	if ( sockit->sock != NULL) sockit->result=mysql_store_result(sockit->sock);

	if ( sockit->result == NULL ) {
		ap_rputs("<TR><TD COLSPAN=50 ALIGN=CENTER>No data to summerize</TD></TR>",r);
	} else {
		comma = mysql_num_rows(sockit->result); 
			/* need result count */

		if ( comma != 0 ) { /* the inner nobody online trap */
		for ( iii=0;iii<comma;iii++) { 
				/* per user loop */
			/* zero out the string */
			memcpy((void *)&ports,"...................................................",50);
			row = mysql_fetch_row(sockit->result); 
				/* get the username */
			if ( row[0] != NULL ) {
			sprintf((char *)&buf,"SELECT DISTINCT NASPortId FROM radacct WHERE TO_DAYS(AcctStartTime) = TO_DAYS(NOW()) AND UserName='%s'",row[0]);
			if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
			if ( sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock);
			if ( sockit->result2 != NULL) {
                             	pct = mysql_num_rows(sockit->result2); 
					/* need result count */
			} else {
				pct=0;
			}
			 for ( ii=0;ii<pct;ii++) { /* per port loop */
				if ( sockit->result2 != NULL) brow = mysql_fetch_row(sockit->result2); 
					/* get the portid */
				if ( sockit->result2 != NULL) {
					if ( brow[0] != NULL) xxx = atoi(brow[0]);
				} else {
					xxx=0;
				}
				if ( xxx < 49 ) ports[xxx] = 'X';
			     } /* end of the per port loop */
			} /* a just in case null trap */
			if ( sockit->result2 != NULL) mysql_free_result(sockit->result2);
			ap_rprintf(r,"<TR><TD><A href=\"%s/data?name=%s\">%s</a></TD>\n",conf->Rurlpath,row[0],row[0]);
			for ( ii=0;ii<49;ii++) {
				ap_rprintf(r,"<TD>%c</TD>",ports[ii]);
				} /* inner port loop */
			ap_rputs("</TR>",r);
                } /* end of the per user loop */
	   } else { /* middle of the nobody online trap */
		ap_rputs("<TR><TD COLSPAN=50 ALIGN=CENTER>No logins to summerize</TD></TR>",r);
	   } /* end of the nobody online trap */
	} /* end of the no users to list trap */
	ap_rputs("<TR><TH>UserName:</TH><TH COLSPAN=49>Ports that this user has used today</TH></TR>",r);
	ap_rputs("</TABLE><BR>",r); /* Done with the table */
	if ( sockit->result != NULL) mysql_free_result(sockit->result);  /* done with this userlist */
return 1;
} /* end of the today instruction */

/* Function to print out login data over a period of days. */
static int rad_todays( rad_cfg *conf, request_rec *r, SQLSOCK *sock) 
{
	int ii,iii,pct,days,comma,xxx;
	struct tm *btime,*today;
	time_t now,then,online, limit;
	MYSQL_ROW row,brow;
        char buf[256], string[128], *p,ports[49], sortby[50];

		/* find out for how many days */
		p = strstr(r->args,"s=");
		p+=2;  /* skip the two characters */
		if ( p == NULL ) {
			days=1;
		} else {
			/* what if the user sends an empty string? */
			if ( *p != (char) 0 ) {
				 days=atoi(p);
			} else {
				 days=1;
			}
		}
		/* start by creating my table */
ap_rputs("<BR><TABLE WIDTH=100% BORDER=1 BGCOLOR=\"ffffff\">",r);
ap_rputs("<TR><TH>UserName:</TH><TH>0<BR>0</TH><TH>0<BR>1</TH><TH>0<BR>2</TH>",r);
ap_rputs("<TH>0<BR>3</TH><TH>0<BR>4</TH><TH>0<BR>5</TH><TH>0<BR>6</TH>",r);
ap_rputs("<TH>0<BR>7</TH><TH>0<BR>8</TH><TH>0<BR>9</TH><TH>1<BR>0</TH>",r);
ap_rputs("<TH>1<BR>1</TH><TH>1<BR>2</TH><TH>1<BR>3</TH><TH>1<BR>4</TH>",r);
ap_rputs("<TH>1<BR>5</TH><TH>1<BR>6</TH><TH>1<BR>7</TH><TH>1<BR>8</TH>",r);
ap_rputs("<TH>1<BR>9</TH><TH>2<BR>0</TH><TH>2<BR>1</TH><TH>2<BR>2</TH>",r);
ap_rputs("<TH>2<BR>3</TH><TH>2<BR>4</TH><TH>2<BR>5</TH><TH>2<BR>6</TH>",r);
ap_rputs("<TH>2<BR>7</TH><TH>2<BR>7</TH><TH>2<BR>9</TH><TH>3<BR>0</TH>",r);
ap_rputs("<TH>3<BR>1</TH><TH>3<BR>2</TH><TH>3<BR>3</TH><TH>3<BR>4</TH>",r);
ap_rputs("<TH>3<BR>5</TH><TH>3<BR>6</TH><TH>3<BR>7</TH><TH>3<BR>8</TH>",r);
ap_rputs("<TH>3<BR>9</TH><TH>4<BR>0</TH><TH>4<BR>1</TH><TH>4<BR>2</TH>",r);
ap_rputs("<TH>4<BR>3</TH><TH>4<BR>4</TH><TH>4<BR>5</TH><TH>4<BR>6</TH>",r);
ap_rputs("<TH>4<BR>7</TH><TH>4<BR>8</TH></TR>",r); 

/* run the query */
	sprintf((char *)&buf,"SELECT DISTINCT UserName FROM radacct WHERE (TO_DAYS(NOW()) - TO_DAYS(AcctStartTime)) <= '%d' ORDER BY UserName",days);
	if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
	if ( sockit->sock != NULL) sockit->result=mysql_store_result(sockit->sock);

 	if ( sockit->result == NULL ) {
		ap_rputs("<TR><TD COLSPAN=50 ALIGN=CENTER>No data to summerize</TD></TR>",r);
	} else {
		comma = mysql_num_rows(sockit->result); 
			/* need result count */
		if ( comma != 0 ) { 
			/* the inner nobody online trap */
		for ( iii=0;iii<comma;iii++) 
		{ /* per user loop */

			/* zero out the string */
			memcpy((void *)&ports,"...................................................",50);
			row = mysql_fetch_row(sockit->result); 
				/* get the username */
			if ( row[0] != NULL ) {
				sprintf((char *)&buf,"SELECT DISTINCT NASPortId FROM radacct WHERE (TO_DAYS(NOW()) - TO_DAYS(AcctStartTime)) <= '%d' AND UserName='%s'",days,row[0]);
			if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
			if ( sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock);
			if ( sockit->result2 != NULL) {
				 pct = mysql_num_rows(sockit->result2); 
			} else {
				pct=0;
			}
			/* need  result count */
			 for ( ii=0;ii<pct;ii++) { /* per port loop */
				if ( sockit->result2 != NULL) brow = mysql_fetch_row(sockit->result2); 
				/* get the portid */
				if ( brow[0] != NULL) xxx = atoi(brow[0]);
				if ( xxx < 49 ) ports[xxx] = 'X';
			     } /* end of the per port loop */
			} /* a just in case null trap */
			if ( sockit->result2 != NULL)mysql_free_result(sockit->result2);
			ap_rprintf(r,"<TR><TD><A href=\"%s/data?name=%s&days=%d\">%s</a></TD>\n",conf->Rurlpath,row[0],days,row[0]);
			for ( ii=0;ii<49;ii++) {
				ap_rprintf(r,"<TD>%c</TD>",ports[ii]);
			} /* inner port loop */
			ap_rputs("</TR>",r);
                } /* end of the per user loop */
	   } else { 
		/* middle of the nobody online trap */
		ap_rputs("<TR><TD COLSPAN=50 ALIGN=CENTER>No logins to summerize</TD></TR>",r);
	   } /* end of the nobody online trap */
	} /* end of the no users to list trap */
	ap_rprintf(r,"<TR><TH>UserName:</TH><TH COLSPAN=49>Ports that this user has used in the past %d days.</TH></TR>",days);
	ap_rputs("</TABLE><BR>",r); /* Done with the table */
	if ( sockit->result != NULL) mysql_free_result(sockit->result);  /* done with this userlist */
return 1;
} /* end of the todays instruction */


/* Function that prints out the data for a user */ 
static int rad_userdata( rad_cfg *conf, request_rec *r, SQLSOCK *sock) 
{
	int days,iii,res,comma;
	struct tm *btime,*today;
	time_t now,then,online, limit;
	MYSQL_ROW row,brow;
        char buf[256], string[128], *p,ports[49], sortby[50];

/* With the GET method, the data is included in the args field.  */

/* find out if the days section was sent */
	p = strstr((const char *)r->args,"s=");
	/* ALWAYS send the days LAST */
	if ( p == NULL ) {
			days=1;
		} else {
			p+=2;  /* skip the two characters */
			/* what if the user sends an empty string? */
			if ( *p != (char) 0 ) {
				 days=atoi(p);
			} else {
				 days=1;
			}
		}
	 p = strstr((const char *)r->args,"&");
	 if ( p != NULL) *p = (char) 0;  
		/* cut the string */
			// r->args.  ( have to strip away the name= )
	 p = strstr((const char *)r->args,"e=");
	  if ( p == NULL ) ap_rputs("No name sent<BR>\n",r);

	  if (( p != NULL) && ( sockit->sock != NULL) ) { 
/* error if null */

	    p+=2;  /* next character is the name....  */ 

/* Check for a user of this name in our records */
	    sprintf((char *)&buf,"SELECT GroupName FROM usergroup WHERE UserName = '%s'",p);
	    if ( sockit->sock   != NULL) mysql_query(sockit->sock,(char *)&buf);
            if ( sockit->sock   != NULL) sockit->result=mysql_store_result(sockit->sock);
	    if ( sockit->result != NULL) {
		comma = mysql_num_rows(sockit->result);
		} else {
		comma=0;
	     } /* trap for the result data */	
	
	if ( comma == 0 ) {
		// Unable to find the user in the group list.
	    sprintf((char *)&buf,"SELECT NASPortId FROM radacct WHERE UserName = '%s'",p);
	    if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
            if ( sockit->sock != NULL) sockit->result=mysql_store_result(sockit->sock);
	    if ( sockit->result != NULL) {
		comma = mysql_num_rows(sockit->result);
		} else {
		comma=0;
	     } /* trap for the result data */	
        } /* end of the second try */ 
	
	if ( comma == 0 ) {
	    ap_rprintf(r,"No such user as %s.<BR><BR>",p);
	    sprintf((char *)&buf,"SELECT DISTINCT UserName FROM usergroup WHERE UserName LIKE '%s%%'",p);
	    if (sockit->sock != NULL ) mysql_query(sockit->sock,(char *)&buf);
            if (sockit->sock != NULL ) sockit->result=mysql_store_result(sockit->sock);
	    if ( sockit->result != NULL) {
		comma = mysql_num_rows(sockit->result);
		} else {
		comma=0;
	     } /* trap for the result data */	
	    ap_rprintf(r,"%d alternatives found",comma);
		if ( comma > 0 ) {
		ap_rprintf(r,"<FORM METHOD=GET ACTION=\"%s/data\">\n",conf->Rurlpath);
		ap_rputs("<SELECT NAME=\"name\" size=1>\n",r);
			for ( iii=0;iii<comma;iii++) {
                 	        if (sockit->result != NULL) row = mysql_fetch_row(sockit->result);
/* Should I test for row==NULL case first? */
				if ( row[0] != NULL ) ap_rprintf(r,"<OPTION>%s\n",row[0]);
                        } /* end of the counting loop */
		ap_rputs("</SELECT>\n",r);
		ap_rputs("<INPUT TYPE=\"submit\" VALUE=\"find\">\n",r);
		ap_rputs("</FORM>\n",r);
		} /* end of the form */
		ap_rputs("</BODY></HTML>\n",r);
	    if ( sockit->result != NULL) mysql_free_result(sockit->result);
	    return 0;  
} /* No Data trap */ 

/* Get the user's group */ 

            if ( sockit->result != NULL ) mysql_free_result(sockit->result);
	    sprintf((char *)&buf,"SELECT GroupName FROM usergroup WHERE UserName = '%s'",p);
	    if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
            if ( sockit->sock != NULL) sockit->result=mysql_store_result(sockit->sock);
            if ( sockit->result != NULL) row = mysql_fetch_row(sockit->result);
	if ( row[0] == NULL ) {
	    sprintf((char *)&buf,"SELECT GroupName FROM usergroup WHERE UserName = 'DEFAULT'");
	    if ( sockit->sock   != NULL) mysql_query(sockit->sock,(char *)&buf);
            if ( sockit->sock   != NULL) sockit->result=mysql_store_result(sockit->sock);
            if ( sockit->result != NULL) row = mysql_fetch_row(sockit->result);
	} /* check for a default group */
		if ( row[0] != NULL ) {
                     ap_rputs("<TABLE ALIGN=CENTER BORDER=1 BGCOLOR=\"#ffffff\">\n",r);
	             ap_rprintf(r,"<TR bgcolor=\"#f6debc\"><TH COLSPAN=2>%s is in group %s</th></tr>\n",p,row[0]); 

// This is the best place to put the group items.
                    ap_rputs("<TR><TH ALIGN=CENTER COLSPAN=2>Check Items</th></tr>\n",r);
		    sprintf((char *)&buf,"SELECT * FROM radgroupcheck WHERE GroupName = '%s'",row[0]);
		    if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
            	     if ( sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock); 
// Got those now.  Print them.
  			if ( sockit->result2 != NULL) {
          			comma = mysql_num_rows(sockit->result2);
			} else {
			 	comma=0;
			}
			for ( iii=0;iii<comma;iii++) {
                 	    if ( sockit->result2 != NULL) brow = mysql_fetch_row(sockit->result2);
		if ( brow != NULL ) {		
				if( !strcasecmp(brow[2],"Daily-Time-Limit") ) {
	if ( brow[3] == NULL ) 
	{
                ap_rputs("<TR><TD COLSPAN=2>Time limit with no value?</td></tr>\n",r);
	} else {
		then = (time_t)atol(brow[3]);
		limit = then;
		today = gmtime((time_t *)&then);
                ap_rprintf(r,"<TR><TD>Daily-Time-Limit</td><TD> %dH %dM %dS</td></tr>\n",today->tm_hour,
			today->tm_min,today->tm_sec);
	}
		} /* end of the brow == null trap */
					} else {
                                    ap_rprintf(r,"<TR><TD>%s</td><TD> %s</td></tr>\n",brow[2],brow[3]);

				} // done converting the time limit
                        } // done looping for reply elements.
                        if ( sockit->result2 != NULL ) mysql_free_result(sockit->result2); 
// Get the reply items now.
                       ap_rputs("<TR><TH ALIGN=CENTER COLSPAN=2>Reply Items</th></tr>\n",r);
		    sprintf((char *)&buf,"SELECT * FROM radgroupreply WHERE GroupName = '%s'",row[0]);
		    if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
            	    if ( sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock); 

// Got those now.  Print them.
			if ( sockit->result2 != NULL ) {
          			comma = mysql_num_rows(sockit->result2);
			} else {
				comma = 0;
			} /* trap on no inner result */
			for ( iii=0;iii<comma;iii++) {
                 	    if ( sockit->result2 != NULL ) brow = mysql_fetch_row(sockit->result2);
                            if ( brow != NULL) ap_rprintf(r,"<TR><TD>%s</td><TD> %s </td></tr>\n",brow[2],brow[3]);
                        } // done looping for reply elements.
                        ap_rprintf(r,"</TABLE><BR>\n");
                        if (sockit->result2 != NULL) mysql_free_result(sockit->result2);
		} else {
			ap_rprintf(r,"%s is in NO group<BR>\n",p);
		} // end of the group name stuff.
          } /* end of the null pointer trap */
          if ( sockit->result != NULL) mysql_free_result(sockit->result);
        sprintf((char *)&buf,"SELECT NASPortId,AcctStartTime,AcctStopTime,AcctSessionTime,AcctTerminateCause,CallingStationID,RadAcctID FROM radacct WHERE UserName = '%s' AND (TO_DAYS(NOW()) - TO_DAYS(AcctStartTime)) <= %d",p,days);
	if (sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
	if (sockit->sock != NULL) sockit->result=mysql_store_result(sockit->sock);
	if ( sockit->result != NULL ) {
		comma = mysql_num_rows(sockit->result);
	} else {
		comma = 0;
	}
	if ( comma == 0 ) {
		ap_rprintf(r,"%s has not been online in the past %d days.<BR>\n",p,days);
	} else { // Create table.

		ap_rputs("<TABLE WIDTH=100\% BORDER=1 BGCOLOR=\"#ffffff\">\n",r); 
// Table header
	ap_rputs("<TR>\n",r);
	ap_rputs("<TH>Port</TH>\n",r);
	ap_rputs("<TH>Login Time</TH>\n",r);
	ap_rputs("<TH>LogOut Time</TH>\n",r);
	ap_rputs("<TH>Online</TH>\n",r);
	ap_rputs("<TH>Logout Cause</TH>\n",r);
        ap_rputs("<TH>Calling From</TH>\n",r);
	ap_rputs("</TR>\n",r); 

// Loop to fill in the table.
if ( sockit->result != NULL) {
	for ( iii=0;iii<comma;iii++) {
                row = mysql_fetch_row(sockit->result);
		ap_rputs("<TR>\n",r);
		ap_rprintf(r,"<TD ALIGN=CENTER>%s</TD>\n",row[0]);
		ap_rprintf(r,"<TD ALIGN=CENTER>%s</TD>\n",row[1]);
		ap_rprintf(r,"<TD ALIGN=CENTER>%s</TD>\n",row[2]); 

// Break this down
	if ( row[3] == NULL ) {
		ap_rputs("<TD ALIGN=CENTER>Still Online</TD>\n",r);
	} else {
		then = (time_t)atol(row[3]);
		if ( then == 0 ) {
			ap_rputs("<TD ALIGN=CENTER>Still Online</TD>\n",r);
		    } else {
			today = gmtime((time_t *)&then);
			ap_rprintf(r,"<TD ALIGN=CENTER>%dH %dM %dS</TD>\n",today->tm_hour,
				today->tm_min,today->tm_sec);
		} 
// inner trap for 0 time used
	}
		ap_rprintf(r,"<TD ALIGN=CENTER>&nbsp <a href=\"%s/showme?%s\">?</a> %s</TD>\n",conf->Rurlpath,row[6],row[4]);
                ap_rprintf(r,"<TD ALIGN=CENTER>&nbsp %s</TD>\n",row[5]);
		ap_rputs("</TR>\n",r);
	} // end of the usage listing. 
		} /* MUST trap null result condition */

// Table footer.
	if ( sockit->result != NULL) mysql_free_result(sockit->result);
	ap_rputs("<TR>\n",r);
	ap_rputs("<TH>Port</TH>\n",r);
	ap_rputs("<TH>Login Time</TH>\n",r);
	ap_rputs("<TH>LogOut Time</TH>\n",r);
	ap_rputs("<TH>Online</TH>\n",r);
	ap_rputs("<TH>Logout Cause</TH>\n",r);
        ap_rputs("<TH>Calling From</TH>\n",r);
	ap_rputs("</TR>\n",r); 

// Now, get the total time logged. 
sprintf((char *)&buf,"SELECT SUM(AcctSessionTime) FROM radacct WHERE UserName = '%s' AND TO_DAYS(AcctStartTime) = TO_DAYS(NOW())",p);
	if ( sockit->sock != NULL) mysql_query(sockit->sock,(char *)&buf);
	if ( sockit->sock != NULL) sockit->result2=mysql_store_result(sockit->sock);
	if (sockit->result2 != NULL ) {
		brow=mysql_fetch_row(sockit->result2);

	if ( brow[0] == NULL ) {
		ap_rputs("<TR>\n",r);
                ap_rputs("<TH COLSPAN=3 ALIGN=RIGHT>Total Used Today</TH>\n",r);
                ap_rputs("<TH ALIGN=CENTER COLSPAN=3> None </TH>\n",r);
		ap_rputs("</TR>\n",r);
	} else {
		then = (time_t)atol(brow[0]);
		if ( (limit > then) && (limit != 0) ) {
			ap_rputs("<TR bgcolor=\"#f6debc\">\n",r);
		} else {
			ap_rputs("<TR bgcolor=\"#ff9999\">\n",r);
		}
	                ap_rputs("<TH COLSPAN=3 ALIGN=RIGHT>Total Used Today</TH>\n",r);
	                today = gmtime((time_t *)&then);
	                ap_rprintf(r,"<TD ALIGN=CENTER COLSPAN=3>%dH %dM %dS</TD>\n",today->tm_hour,
	                        today->tm_min,today->tm_sec);
			ap_rputs("</TR>\n",r); 

/* now list the lime limit for the day */
		if ( limit > 0 ) {
			ap_rputs("<TR bgcolor=\"#f6debc\">\n",r);
			today = gmtime((time_t *)&limit);
			ap_rprintf(r,"<TH COLSPAN=3>Daily-Time-Limit</TH>\n<TD COLSPAN=3 ALIGN=CENTER>%dH %dM %dS</TD>\n",today->tm_hour,
				today->tm_min,today->tm_sec);
		}
		ap_rputs("</TR>\n",r);
	} /* add up the total time used */
	} /* end of the brow==null trap */ 
/* At this point I want to give a breakdown for the past 30 days. */
		if ( sockit->result2 != NULL) mysql_free_result(sockit->result2);
			ap_rputs("<TR><th colspan=6>Login history for the past 30 days</TH></TR>\n",r);
for ( iii=1;iii<31;iii++) {
			if ( iii == 1 ) {
				ap_rputs("<TR><th colspan=3>Yesterday</TH>\n",r);
			} else {
				ap_rprintf(r,"<tr><TH COLSPAN=3>%d days ago</TH>\n",iii);
			} /* neatness for the display of days vs. day */ 
	sprintf((char *)&buf,"SELECT SUM(AcctSessionTime) FROM radacct WHERE UserName = '%s' AND TO_DAYS(AcctStartTime) = (TO_DAYS(NOW())-%d)",p,iii);
	/* Run the query */
	if ( sockit->sock != NULL)   mysql_query(sockit->sock,(char *)&buf);
	if ( sockit->sock != NULL)   sockit->result2=mysql_store_result(sockit->sock);
	if ( sockit->result2 != NULL) { 
		brow=mysql_fetch_row(sockit->result2);
	} else {
		brow = NULL;
	}
	if ( brow != NULL ) {
		if (brow[0] != NULL) {
			then = (time_t)atol(brow[0]);
 	        	today = gmtime((time_t *)&then);
 		        ap_rprintf(r,"<TD ALIGN=CENTER COLSPAN=3>%dH %dM %dS</TD>\n",today->tm_hour,
        	                today->tm_min,today->tm_sec);
			ap_rputs("</TR>\n",r);
		} else { /* cap for NULL TRAP */
			ap_rputs("<td align=center colspan=3>-----</td></tr>\n",r);
		} /* No data display */
	} else { /* main NULL trap */
			ap_rputs("<td align=center colspan=3>-----</td></tr>\n",r);
        }
		if ( sockit->result2 != NULL) mysql_free_result(sockit->result2);
} /* done counting to 30 */

		ap_rputs("</Table>\n<BR>",r);
	} // end of the user was online table.

return 1;
} /* end of user data printing function */


/* this routine just prints everything out */ 
static int rad_debug( rad_cfg *conf, request_rec *r) 
{
	char *p,*pp,*ppp;

ap_rputs("<table width=75% border=1>\n",r); 
ap_rputs("<tr><th colspan=2>Directives</th></tr>\n",r); 
ap_rprintf(r,"<tr><th>Auth File</th><td>%s</td></tr>\n",conf->RauthFile); 
ap_rprintf(r,"<tr><th>DB Pass</th><td>%s</td></tr>\n",conf->Rdbpass); 
ap_rprintf(r,"<tr><th>DB User</th><td>%s</td></tr>\n",conf->Rdbuser); 
ap_rprintf(r,"<tr><th>Database</th><td>%s</td></tr>\n",conf->Rdatabase); 
ap_rprintf(r,"<tr><th>URI</th><td>%s</td></tr>\n",conf->Rurlpath); 
ap_rprintf(r,"<tr><th>DB server</th><td>%s</td></tr>\n",conf->Rdbserver); 
ap_rprintf(r,"<tr><th>Highlighting</th><td>%s</td></tr>\n",conf->Rhcolor);

ap_rputs("<tr><th colspan=2>The Request</th></tr>\n",r); 
ap_rprintf(r,"<tr><th>The Request</th><td>%s</td></tr>\n",r->the_request); 
ap_rprintf(r,"<tr><th>Protocol</th><td>%s</td></tr>\n",r->protocol); 
ap_rprintf(r,"<tr><th>Host Name</th><td>%s</td></tr>\n",r->hostname); 
ap_rprintf(r,"<tr><th>Status</th><td>%s</td></tr>\n",r->status_line); 
ap_rprintf(r,"<tr><th>Method</th><td>%s</td></tr>\n",r->method); 
ap_rprintf(r,"<tr><th>Type</th><td>%s</td></tr>\n",r->content_type); 
ap_rprintf(r,"<tr><th>Handler</th><td>%s</td></tr>\n",r->handler); 
ap_rprintf(r,"<tr><th>Encoding</th><td>%s</td></tr>\n",r->content_encoding); 
ap_rprintf(r,"<tr><th>Validator</th><td>%s</td></tr>\n",r->vlist_validator); 
ap_rprintf(r,"<tr><th>User</th><td>%s</td></tr>\n",r->connection->user); 
//ap_rprintf(r,"<tr><th>Auth Type</th><td>%s</td></tr>\n",r->ap_auth_type); 
ap_rprintf(r,"<tr><th>Raw</th><td>%s</td></tr>\n",r->unparsed_uri); 
ap_rprintf(r,"<tr><th>URI</th><td>%s</td></tr>\n",r->uri); 
ap_rprintf(r,"<tr><th>Filename</th><td>%s</td></tr>\n",r->filename); 
ap_rprintf(r,"<tr><th>Path Info</th><td>%s</td></tr>\n",r->path_info); 
ap_rprintf(r,"<tr><th>Args</th><td>%s</td></tr>\n",r->args);

//ap_rputs("<tr><th colspan=2>Headers</th></tr>\n",r); 
//apr_table_do( writeout_func, r, r->headers_in, NULL );

//ap_rputs("<tr><th colspan=2>Sub Environment</th></tr>\n",r); 
//apr_table_do( writeout_func, r, r->subprocess_env, NULL );

ap_rputs("<tr><th colspan=2>Environment</th></tr>\n",r); 
ap_rprintf(r,"<tr><th colspan=2>%s</td></tr>\n",environ);

if ( r->args != NULL ) {

pp=NULL;
ppp=NULL;

ap_rputs("<tr><th colspan=2>Breakdown</th></tr>\n",r); 
	p=r->args;
if ( p  != NULL ) pp=strstr(p,"=");
if ( p  != NULL ) ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( p  != NULL ) ap_rprintf(r,"<tr><td>%s</td>",p);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);
if ( ppp != NULL ) pp=strstr(ppp,"=");
if ( ppp != NULL ) p =strstr(ppp,"&");
if ( p != NULL ) *p=(char) 0;
if ( p != NULL ) p++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( ppp != NULL ) ap_rprintf(r,"<tr><td>%s</td>",ppp);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);
if ( p  != NULL ) pp=strstr(p,"=");
if ( p  != NULL ) ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( p  != NULL ) ap_rprintf(r,"<tr><td>%s</td>",p);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);
if ( ppp != NULL ) pp=strstr(ppp,"=");
if ( ppp != NULL ) p =strstr(ppp,"&");
if ( p != NULL ) *p=(char) 0;
if ( p != NULL ) p++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( ppp  != NULL ) ap_rprintf(r,"<tr><td>%s</td>",ppp);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);
if ( p  != NULL ) pp=strstr(p,"=");
if ( p  != NULL ) ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( p  != NULL ) ap_rprintf(r,"<tr><td>%s</td>",p);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);
if ( ppp != NULL ) pp=strstr(ppp,"=");
if ( ppp != NULL ) p =strstr(ppp,"&");
if ( p != NULL ) *p=(char) 0;
if ( p != NULL ) p++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( ppp  != NULL ) ap_rprintf(r,"<tr><td>%s</td>",ppp);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);
if ( p  != NULL ) pp=strstr(p,"=");
if ( p  != NULL ) ppp=strstr(p,"&");
if ( ppp != NULL ) *ppp=(char) 0;
if ( ppp != NULL ) ppp++;
if ( pp != NULL ) *pp=(char) 0;
if ( pp != NULL ) pp++;
if ( p  != NULL ) ap_rprintf(r,"<tr><td>%s</td>",p);
if ( pp != NULL ) ap_rprintf(r,"<td>%s</td></tr>",pp);

} /* Block off the NULL r->args case */

/* close up the table and go home */ 
ap_rputs("</table>\n\n<br><br>\n",r);

 return 0; 
}


static char *get_Rauth_pw ( request_rec *r, char *user, char *Rauthfile ) 
{
	char *pw=NULL;
	int res;  /* results */
	GDBM_FILE dbf;  /* Auth File itself */
	datum key,dat;  /* data for lookup */

    if ( user == NULL ) {
	//ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"GDBM username not sent to validate %s",Rauthfile);
	syslog(LOG_ERR,"GDBM username not sent to validate %s",Rauthfile);
	return NULL;
    }

/* open the database */ 
dbf=gdbm_open(Rauthfile,512,GDBM_READER,622,0);

/* file failed to open */
  if ( dbf==NULL ) {
//	ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r, "GDBM file %s would not open. %s",Rauthfile,user);
	syslog(LOG_ERR, "GDBM file %s would not open. %s",Rauthfile,user);
   return NULL;
  }

/* what are we looking for? */

key.dptr=user; key.dsize=strlen(user);

/* is this a real user name ? */ 
res=gdbm_exists(dbf,key);

 if (!res) {
//	ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r, "No such user as %s",user);
	syslog(LOG_ERR, "No such user as %s",user);
   return NULL;
  }

/* username exists if we get here */

dat=gdbm_fetch( dbf,key );

/* no such username */
  if (dat.dptr==NULL ) {
//	ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r, "GDBM file %s did not find %s",Rauthfile,user);
	syslog(LOG_ERR, "GDBM file %s did not find %s",Rauthfile,user);
   return NULL;
  }

/* null terminate the result */ 
dat.dptr[dat.dsize]=0;

/* close the database */ 
gdbm_close(dbf);

/* set the pointer */ 
pw=(char *)dat.dptr;

/* free(pw); / * MUST be manually freed */

	/* if we get a password, return it. */ 
return (char *)pw; 
}

/* --------------------------------------------------------
 *
 * Now for the exposed routines
 *
 */

/* Actually check authorization now. */
static int Radius_gdbm_authenticate(request_rec *r)
{
rad_cfg *conf = ap_get_module_config(r->per_dir_config,	&radwho_module);
char *user = r->connection->user;
int  res;
const char *sent_pw;
char *real_pw;
GDBM_FILE dbf;
datum key,username;


syslog(LOG_INFO,"Attempting to locate User");

if ((res = ap_get_basic_auth_pw(r,&sent_pw))) return res;

/* on deny, return DECLINED    */
/* return AUTH_REQUIRED to ask */
/* return OK to permit access  */

user = r->connection->user;

if (user == NULL ) {
	syslog(LOG_DEBUG,"No username sent %s",sent_pw);
        return AUTH_REQUIRED;
}
/* open the datadabse and get my password. */
	/* We have a database--right? */
	if ( !conf->RauthFile )
	{
//	ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,"GDBM authentication file MUST be defined.");
	syslog(LOG_ERR,"GDBM authentication file MUST be defined.");
		return DECLINED;
	}
/* ask for the password from the database */
	if (!(real_pw=get_Rauth_pw(r,r->connection->user,conf->RauthFile)))
	{
	/* I'm always authoritative */
	 return DECLINED;
	}
	if (!strncmp(sent_pw,real_pw,strlen(real_pw)) ) {
			syslog(LOG_INFO,"%s login successful",r->connection->user);
			free(real_pw); 
 			return OK; 
	}
 
free(real_pw); 
return DECLINED;  /* permit access. */
}
/* End of the authorization check. */

static int Radius_gdbm_locater ( request_rec *r ) 
{
	rad_cfg *conf = ap_get_module_config(r->per_dir_config,	&radwho_module);

	int res;  /* results */
	GDBM_FILE dbf;  /* Auth File itself */
	datum key;  /* data for lookup */

syslog(LOG_INFO,"Attempting to locate User IN database");


if ( r->connection->user == NULL ) {
	syslog(LOG_INFO,"Username NOT specified yet.");
	return DECLINED;
}

if ( conf == NULL ) {
	syslog(LOG_ERR,"No configuration data");
	return DECLINED;
}

if ( conf->RauthFile == NULL ) {
	syslog(LOG_ERR,"No Authentication File data");
	return DECLINED;
}

syslog(LOG_INFO,"Opening database");

/* open the database */ 
dbf=gdbm_open(conf->RauthFile ,512,GDBM_READER,622,0);

/* file failed to open */
  if ( dbf==NULL ) {
//	ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r, "GDBM file %s would not open.",conf->RauthFile);
	syslog(LOG_ERR, "GDBM file %s would not open.",conf->RauthFile);
   return DECLINED;
  }

/* what are we looking for? */
/* the r structure has no user data */
key.dptr=r->connection->user; 
key.dsize=strlen(r->connection->user);

syslog(LOG_INFO,"Attempting to locate %s IN database",r->connection->user);

/* is this a real user name ? */ 

res=gdbm_exists(dbf,key);

 if (!res) {
	syslog(LOG_ERR, "No such user as %s",r->connection->user);
   return DECLINED;
  }

/* username exists if we get here */
		syslog(LOG_INFO,"User %s exists.",r->connection->user);

/* close the database */
 gdbm_close(dbf);
 return OK; 
}

static int radwho_handler(request_rec *r) 
{
	rad_cfg *conf; 
	int usercount,iii,res,days,extra;
	struct tm *btime,*today;
	time_t now,then,online, limit;
	MYSQL_ROW row,brow;
        char buf[256], string[128], *p,ports[49], sortby[50]; 


/* keep the module from answering ALL requests. */
	if (strcmp(r->handler,"radwho") ) return DECLINED;

	syslog(LOG_INFO,"Begin Request Handler");
	if ( r == NULL ) return DECLINED;

	syslog(LOG_INFO,"Get config record");

	conf = ap_get_module_config(r->per_dir_config, &radwho_module);


if ( conf == NULL ) {
	syslog(LOG_ERR,"NULL configuration for this directory.");
	return DECLINED;
}

/*
 * Let's load up the sortby variable
 * Information can come from any of 3 places.
 * 1) the RADSortBy directive
 * 2) the default value of NASPortId
 * 3) the request itself.
 * I'll set it to each, in turn, and leave the last one in place.
 */


	syslog(LOG_INFO,"Get memory for socket...");
/*	sockit=apr_palloc(r->pool, sizeof(SQLSOCK));  */
	sockit=malloc(sizeof(SQLSOCK)); 
	syslog(LOG_INFO,"Socket Memory Set.");



snprintf((char *)&sortby,49,"NASPortId");
 
if ( conf->RSortBy != NULL) snprintf((char *)&sortby,49,"%s",conf->RSortBy);

	syslog(LOG_INFO,"sortby %s",sortby);

if ( ( p=strstr(r->path_info,"/sort") ) != NULL) 
{

	/* Set the value */
	p=strstr(p,"=");
	if ( p != NULL ) snprintf((char *)&sortby,49,"%s",p);

	/* trim the string so we have no junk to work with */
	p=strstr((char *)&sortby,"&");
	if ( p != NULL ) *p=(char) 0 ;
	p=strstr((char *)&sortby,"+");
	if ( p != NULL ) *p=(char) 0 ;
	p=strstr((char *)&sortby,"\n");
	if ( p != NULL ) *p=(char) 0 ;
	p=strstr((char *)&sortby,"\r");
	if ( p != NULL ) *p=(char) 0 ;
	p=strstr((char *)&sortby,"%");
	if ( p != NULL ) *p=(char) 0 ; 
}



/* r->path_info should contain my keys */
	syslog(LOG_INFO,"Begin Header...");

r->content_type= "text/html"; 
ap_send_http_header(r);
if (r->header_only) return OK;

/* debug line */ 
usercount=0;


/* construct the document to send */ 
ap_rputs(DOCTYPE_HTML_3_2,r); 
ap_rputs("<html>\n<head>\n",r); 
ap_rputs("<title>Radius Information</title>\n",r); 
ap_rputs("<meta http-equiv=\"Refresh\" content=\"300\">\n",r); 
ap_rputs("</head>\n<body bgcolor=\"#33ccff\">\n",r); 
ap_rputs("<center><h1>Icradius Login Information</h1></center>\n",r);

	syslog(LOG_INFO,"Begin Table...");

/* this is where I'll be putting the data. */

/* Create the table under all conditions. */
/* put the header on the table */ 
ap_rputs("<table width=100% border=1 bgcolor=\"#ffffff\">\n",r);

/* I use the current time below even if the socket fails. */
now=time(NULL); 

	syslog(LOG_INFO,"Got Time...");


if ( sockit == NULL ) {
	ap_rputs("<h1>ERROR</h1>No SQL socket memory available<br>\n",r); 
} else {

/* Have a socket, lets try to connect and get some data... */

/* ap_hard_timeout called here here */
    ap_hard_timeout("send radius online", r);

	syslog(LOG_INFO,"Init Socket..");
if ( &(sockit->conn) == NULL ) {
	syslog(LOG_INFO,"Socket Connection is NULL..");
}


/* initialize the socket */ 
mysql_init(&(sockit->conn));

	syslog(LOG_INFO,"Connection Attempt");

/* try to connect */ 
sockit->sock=mysql_real_connect(&(sockit->conn),conf->Rdbserver,conf->Rdbuser,conf->Rdbpass,conf->Rdatabase,0,NULL,0);

/*	syslog(LOG_INFO,"Connected"); */

if (sockit->sock == NULL ) {
	syslog(LOG_ERR,"Unable to connect to mysql server %s as %s [%s]",conf->Rdbserver,conf->Rdbuser,mysql_error(&(sockit->conn)) );
	ap_rputs("<tr><th colspan=7 bgcolor=\"#f6debc\">\n",r); 
	ap_rputs("<b>Error:</b>Connection failed</th></tr>\n",r);
	ap_rputs("</th></tr>\n<tr><th colspan=7>\n",r); now=time(NULL); 
	ap_rprintf(r,"Server Time: %s </th></tr>\n",ctime((time_t *)&now));
	usercount=-1; 
	} else { 
	syslog(LOG_INFO,"Connected"); 
/* Have to clear the data first.... */
if (!strncmp(r->path_info,"/clear",6)) rad_clear( conf, r, sockit);


	/* get the data to print out */ 
	syslog(LOG_INFO,"Creating query for socket.");
		sprintf((char *)&string,"SELECT RadAcctID,UserName,NASPortId,AcctStartTime,FramedIPAddress,UNIX_TIMESTAMP(AcctStartTime),ConnectInfo from radacct where AcctSessionTime='0' ORDER BY '%s'",sortby); 
		if ( sockit->sock != NULL ) mysql_query(sockit->sock,(char *)&string); 
		if ( sockit->sock != NULL ) sockit->result=mysql_store_result(sockit->sock);
	if ( sockit->result != NULL ){
	syslog(LOG_INFO,"Got usercount.");
		usercount=mysql_num_rows(sockit->result);
	} else {
		usercount=0; 
	}

/* have to get the login count before reaching here */ 
ap_rputs("<tr><th colspan=7 bgcolor=\"#f6debc\">\n",r); 
if ( usercount == 0 ) ap_rprintf(r,"No users online ( sorted by %s ).\n",usercount, sortby); 
if ( usercount == 1 ) ap_rprintf(r,"1 User are online at this time ( sorted by %s ).\n", sortby); 
if ( usercount > 1 ) ap_rprintf(r,"%d Users are online at this time ( sorted by %s ).\n",usercount, sortby); 

ap_rputs("</th></tr>\n<tr><th colspan=7>\n",r); 
ap_rprintf(r,"Server Time: %s </th></tr>\n",ctime((time_t *)&now));

/* now we put a header on the table */
	ap_rputs("<TR>\n",r);  /* Header repeated below for clarity */
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=UserName\">UserName</a></TH>\n",conf->Rurlpath);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=NASPortId\">Port</a></TH>\n",conf->Rurlpath);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=AcctStartTime\">Login Time</a></TH>\n",conf->Rurlpath);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=FramedIPAddress\">User IP</a></TH>\n",conf->Rurlpath);
        ap_rputs("<TH>Online</TH>\n",r);
	ap_rputs("<TH>Today</TH>\n",r);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=ConnectInfo\">Info</a></TH>\n",conf->Rurlpath);
	ap_rputs("</TR>\n",r);

/* don't print on error or nobody online.

if ( usercount > 0 ) {

/* loop until usercount */
	for ( iii=0;iii<usercount;iii++) {
		row=mysql_fetch_row(sockit->result);

	/* have the data, I can begin the printout */ 
	
/* get the login time */
	then=(time_t)atol(row[5]);
	online = now - then;

/* convert and break out fields. */
	btime=gmtime((time_t *)&online);

/* Need the total online time for this day */ 
sprintf((char *)&string,"SELECT SUM(AcctSessionTime) FROM radacct where UserName='%s' AND TO_DAYS(AcctStartTime) = TO_DAYS(NOW())",row[1]);
	if ( sockit->sock != NULL ) mysql_query(sockit->sock,(char *)&string);
	if ( sockit->sock != NULL ) sockit->result2=mysql_store_result(sockit->sock);
	if ( sockit->result2 != NULL) {
		brow=mysql_fetch_row(sockit->result2);
		if ( brow[0] == NULL ) {
			then = online;
		} else {
			then = (time_t)atol(brow[0]);
			then += online;
		} /* add the current time used to the total. */
	} else {
			then = online;
		ap_rprintf(r,"<tr><th colspan=7>Unable to fetch login data for %s</th></tr>\n",row[1]);
	} /* hard way--there was no result to work with */

/* get the usage cap flag and set the limit */
	if ( conf->RLimit != NULL) {
		extra=(int)atoi(conf->RLimit);
	} else {
		extra=35999;
	}

if ( then > extra ) 
{ 
/* in excess of X hours. -1 second [default] 35999 */
		if ( conf->Rhcolor == NULL ) {
			ap_rputs("<TR bgcolor=\"#ff9999\">\n",r);
		} else {
			ap_rprintf(r,"<TR bgcolor=\"%s\">\n",conf->Rhcolor);
		}
	} else {
		ap_rputs("<TR>\n",r); 
        } 

/* end of the usage cap */

        ap_rprintf(r,"<TD ALIGN=CENTER><A HREF=\"%s/data?name=%s\">%s</A></TD>\n",conf->Rurlpath,row[1],row[1]);
	ap_rprintf(r,"<TD ALIGN=CENTER>%s</TD>\n",row[2]);
	ap_rprintf(r,"<TD ALIGN=CENTER>%s</TD>\n",row[3]);
	ap_rprintf(r,"<TD ALIGN=CENTER>%s</TD>\n",row[4]);
	ap_rprintf(r,"<TD ALIGN=CENTER>%dH %dM %dS</TD>\n",btime->tm_hour,btime->tm_min,btime->tm_sec); 

 /* Since the space used to store this structure is overwritten
    each time gmtime is called, I cannot call this before writing
    out the earlier data.  Otherwise, this data gets printed in both places. 
*/

	today = gmtime((time_t *)&then);
	ap_rprintf(r,"<TD ALIGN=CENTER>%dH %dM %dS</TD>\n",today->tm_hour,today->tm_min,today->tm_sec);
	ap_rprintf(r,"<TD><a href=\"%s/clear?%s\">X</a>&nbsp <a href=\"%s/showme?%s\">?</a> %s</TD>\n",conf->Rurlpath,row[0],conf->Rurlpath,row[0],row[6]);
	ap_rputs("</TR>\n",r);
	if ( sockit->result2 != NULL) mysql_free_result(sockit->result2);

	} /* end of the counting loop */ 
} /* the NO users online trap */

/*   Original header retained for reference.  
	ap_rputs("<tr>\n",r);
	ap_rputs("<th>UserName</th>\n",r);
	ap_rputs("<th>Port</th>\n",r);
	ap_rputs("<th>LoginTime</th>\n",r);
	ap_rputs("<th>UserIP</th>\n",r);
	ap_rputs("<th>Online</th>\n",r);
	ap_rputs("<th>Today</th>\n",r);
	ap_rputs("<th>Info</th>\n",r);
	ap_rputs("</tr>\n",r); 
*/


/* Repeat the header at the bottom of the table. */ 

	ap_rputs("<TR>\n",r);  
/* repeat the header */
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=UserName\">UserName</a></TH>\n",conf->Rurlpath);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=NASPortId\">Port</a></TH>\n",conf->Rurlpath);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=AcctStartTime\">Login Time</a></TH>\n",conf->Rurlpath);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=FramedIPAddress\">User IP</a></TH>\n",conf->Rurlpath);
        ap_rputs("<TH>Online</TH>\n",r);
	ap_rputs("<TH>Today</TH>\n",r);
        ap_rprintf(r,"<TH><a href=\"%s/sort?by=ConnectInfo\">Info</a></TH>\n",conf->Rurlpath);
	ap_rputs("</TR>\n",r);

/* Before we close the table, we create the form to get more data on a single user.... */
	ap_rputs("<TR><TH COLSPAN=3 bgcolor=\"#f6debc\">\n",r);
	ap_rprintf(r,"<FORM METHOD=GET ACTION=\"%s/data\">\n",conf->Rurlpath);
	ap_rputs("<INPUT NAME=\"name\" size=15>\n",r);
	ap_rputs("<INPUT TYPE=\"submit\" VALUE=\"find\">\n",r);
	ap_rputs("</FORM></TH>\n",r);
	ap_rprintf(r,"<TH COLSPAN=1 bgcolor=\"#f6debc\"><FORM METHOD=GET ACTION=\"%s/today\">\n",conf->Rurlpath);
	ap_rputs("<INPUT TYPE=\"submit\" VALUE=\"Today\">\n",r);
	ap_rputs("</FORM></TH>\n<TH COLSPAN=3 bgcolor=\"#f6debc\">\n",r);
	ap_rprintf(r,"<FORM METHOD=GET ACTION=\"%s/todays\">\n",conf->Rurlpath);
	ap_rputs("<INPUT NAME=\"days\" size=2 value=\"1\">\n",r);
	ap_rputs("<INPUT TYPE=\"submit\" VALUE=\"days\">\n",r);
	ap_rputs("</FORM></TD></TR>\n",r);


	ap_rputs("<TR><TH COLSPAN=7 bgcolor=\"#bcdefc\">\n",r);

	today = localtime((time_t *)&now);

	ap_rprintf(r,"<FORM METHOD=GET ACTION=\"%s/whohad\">\n",conf->Rurlpath);
	ap_rputs("Who Had:<INPUT NAME=\"ip\" size=15 value=\"127.000.000.001\">\n",r);
	ap_rprintf(r,"On:<INPUT NAME=\"year\" size=4 value=\"%d\">\n",today->tm_year+1900);
	ap_rprintf(r,"-<INPUT NAME=\"month\" size=2 value=\"%d\">\n",today->tm_mon+1);
	ap_rprintf(r,"-<INPUT NAME=\"day\" size=2 value=\"%d\">\n",today->tm_mday);
	ap_rprintf(r,"At:<INPUT NAME=\"hr\" size=2 value=\"%d\">\n",today->tm_hour);
	ap_rprintf(r,":<INPUT NAME=\"min\" size=2 value=\"%d\">\n",today->tm_min);
	ap_rprintf(r,":<INPUT NAME=\"sec\" size=2 value=\"%d\">\n",today->tm_sec);
	ap_rputs("<INPUT TYPE=\"submit\" VALUE=\"Seek\">\n",r);
	ap_rputs("</FORM>\n</th></tr>\n",r);


	/* Finish off the table */ 
	ap_rputs("\n</table><br><br><br>\n",r);


/* Print the additional funtion output below the table */

if (!strncmp(r->path_info,"/todays",7)) rad_todays( conf, r, sockit);
if (!strncmp(r->path_info,"/whohad",7)) rad_whohad( conf, r, sockit);
if (!strncmp(r->path_info,"/showme",7)) rad_showme( conf, r, sockit);
if (!strncmp(r->path_info,"/today",6)) rad_today( conf, r, sockit);
if (!strncmp(r->path_info,"/data",5)) rad_userdata( conf, r, sockit);

/* close the connection if we have one */ 
if ( usercount > -1 ) {
	if (sockit->result != NULL) mysql_free_result(sockit->result);
	if (sockit->sock != NULL) mysql_close(sockit->sock); 

}

 if ( sockit != NULL) free(sockit);

} /* end of the no socket condition. */

if (!strncmp(r->path_info,"/debug",6)) rad_debug( conf, r); 


/* Print out the usage information */
	ap_rputs("<u>Usage:</u><br>\n", r);
	ap_rputs("Click the table headings to sort the listings.<br>\n", r);
	ap_rputs("The 'X' in the info column deletes a dead login.<br>\n", r);
	ap_rputs("The '?' in the info column shows the entire login record for a login.<br>\n", r);
        ap_rputs("Click the Username to get todays logins for the user.<br>\n", r);
	ap_rputs("<br>\n", r);  /* End of the instructions */

    ap_rputs("Mod_Radwho Version 3.0<br>\n", r);

/* report success */ 
ap_rputs("\n</body></html>",r);
return OK; 
}

/* These commands are valid within Directory,Location and Files sections */
/* Version 1.3 comand structure */
static const command_rec radwho_cmds[] ={

/* FIXED below here */
	{"RADAuthFile",set_rad_authfile,NULL,ACCESS_CONF,TAKE1,
	"gdbm file with username/password pairs."},

	{"RADUserLimit",set_rad_limitflag,NULL,ACCESS_CONF,TAKE1,
	"Time limit, in seconds, when a user login if flagged."},

	{"RADSortBy",set_rad_sortby,NULL,ACCESS_CONF,TAKE1,
	"Default sorting field."},

	{"RADDBPass",set_rad_dbpass,NULL,ACCESS_CONF,TAKE1,
	"Login Password for the MYSQL database."},

	{"RADDBUser",set_rad_dbuser,NULL,ACCESS_CONF,TAKE1,
	"Login Username for the MYSQL database."},

	{"RADServer",set_rad_dbserver,NULL,ACCESS_CONF,TAKE1,
	"Server with the radius database on it."},

	{"RADDatabase",set_rad_database,NULL,ACCESS_CONF,TAKE1,
	"Database containing login data."},
	
	{"RADURL",set_rad_urlpath,NULL,ACCESS_CONF,TAKE1,
	"URL of this page on the server."},

	{"RADHighlight",set_rad_highlight,NULL,ACCESS_CONF,TAKE1,
	"Color code for highlighting over-time users."},
{NULL}
};

/* Version 1.3 handler structure */
static const handler_rec radwho_handlers[] =
{
    {"radwho", radwho_handler},
    {NULL}
};

/* original from version 1.3 */
module MODULE_VAR_EXPORT radwho_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                                  /* initializer */
    create_radwho_config_record,           /* dir config creater */
    NULL,                                  /* dir merger --- default is to override */
    NULL,                                  /* server config */
    NULL,                                  /* merge server config */
    radwho_cmds,                           /* command table */
    radwho_handlers,            	   /* handlers */
    NULL,                   	           /* filename translation */
    Radius_gdbm_authenticate,              /* check_user_id */
    Radius_gdbm_locater,                   /* check auth */
    Radius_gdbm_locater,                   /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL                        /* post read-request */
};

